from cvfinetune.parser.base import FineTuneParser
from cvfinetune.parser.base import add_dataset_args
from cvfinetune.parser.base import add_model_args
from cvfinetune.parser.base import add_training_args
from cvfinetune.parser.base import default_factory


__all__ = [
	"add_dataset_args",
	"add_model_args",
	"add_training_args",
	"default_factory",
	"FineTuneParser",
]




import abc

from cvargparse import Arg
from cvfinetune.parser.utils import DEFAULT_INFO_FILE
from cvfinetune.parser.utils import get_info_file
from cvfinetune.parser.utils import parser_extender

@parser_extender
def add_dataset_args(parser):

	info_file = get_info_file()

	if info_file is None:
		_args = [
			Arg("data"),
			Arg("dataset"),
			Arg("parts")]
	else:
		_args = [
			Arg("data", default=DEFAULT_INFO_FILE),
			Arg("dataset", choices=info_file.DATASETS.keys()),
			Arg("parts", choices=info_file.PART_TYPES.keys()),
		]

	_args.extend([

		Arg("--label_shift", type=int, default=1,
			help="label shift"),

		Arg("--swap_channels", action="store_true",
			help="preprocessing option: swap channels from RGB to BGR"),

		Arg("--n_jobs", "-j", type=int, default=0,
			help="number of loading processes. If 0, then images are loaded in the same process"),

	])

	parser.add_args(_args, group_name="Dataset arguments")

class DatasetParserMixin(abc.ABC):
	def __init__(self, *args, **kwargs):
		super(DatasetParserMixin, self).__init__(*args, **kwargs)
		add_dataset_args(self)


__all__ = [
	"DatasetParserMixin",
	"add_dataset_args"
]



from .sacred import SacredTrainer
from .base import _is_adam, default_intervals

from chainer.training import extensions

def observe_alpha(trainer):
	model = trainer.updater.get_optimizer("main").target.model
	return float(model.pool.alpha.array)

class AlphaPoolingTrainer(SacredTrainer):

	@property
	def model(self):
		return self.updater.get_optimizer("main").target.model

	def __init__(self, opts, updater, *args, **kwargs):
		super(AlphaPoolingTrainer, self).__init__(opts=opts, updater=updater, *args, **kwargs)
		### Alternating training of CNN and FC layers (only for alpha-pooling) ###
		if opts.switch_epochs:
			self.extend(SwitchTrainables(
				opts.switch_epochs,
				model=self.model,
				pooling=self.model.pool))

	def reportables(self, opts):
		print_values, plot_values = super(AlphaPoolingTrainer, self).reportables()
		alpha_update_rule = self.model.pool.alpha.update_rule
		if _is_adam(opts):
			# in case of Adam optimizer
			alpha_update_rule.hyperparam.alpha *= opts.kappa
		else:
			alpha_update_rule.hyperparam.lr *= opts.kappa

		self.extend(
			extensions.observe_value("alpha", observe_alpha),
			trigger=default_intervals.print)

		print_values.append("alpha")
		plot_values["alpha"]= ["alpha"]

		return print_values, plot_values



import numpy as np
import abc

from chainer_addons.dataset import AugmentationMixin
from chainer_addons.dataset import PreprocessMixin

from cvdatasets.dataset import AnnotationsReadMixin
from cvdatasets.dataset import RevealedPartMixin
from cvdatasets.dataset import IteratorMixin

class _pre_augmentation_mixin(abc.ABC):
	""" This mixin discards the parts from the ImageWrapper object
	and shifts the labels
	"""

	label_shift = 1

	def get_example(self, i):
		im_obj = super(_pre_augmentation_mixin, self).get_example(i)
		im, parts, lab = im_obj.as_tuple()
		return im, lab + self.label_shift

class _base_mixin(abc.ABC):
	""" This mixin converts images,that are in range
	[0..1] to the range [-1..1]
	"""

	def get_example(self, i):
		im, lab = super(_base_mixin, self).get_example(i)

		if isinstance(im, list):
			im = np.array(im)

		if np.logical_and(0 <= im, im <= 1).all():
			im = im * 2 -1

		return im, lab


class BaseDataset(_base_mixin,
	# augmentation and preprocessing
	AugmentationMixin, PreprocessMixin,
	_pre_augmentation_mixin,
	# random uniform region selection
	RevealedPartMixin,
	# reads image
	AnnotationsReadMixin,
	IteratorMixin):
	"""Commonly used dataset constellation"""



import abc

from cvargparse import Arg
from cvargparse import ArgFactory
from cvfinetune.parser.utils import parser_extender

from chainer_addons.training import OptimizerType

@parser_extender
def add_training_args(parser):

	_args = ArgFactory([

		Arg("--warm_up", type=int, help="warm up epochs"),

		OptimizerType.as_arg("optimizer", "opt",
			help_text="type of the optimizer"),

		Arg("--cosine_schedule", type=int,
			default=-1,
			help="enable cosine annealing LR schedule. This parameter sets the number of schedule stages"),

		Arg("--l1_loss", action="store_true",
			help="(only with \"--only_head\" option!) use L1 Hinge Loss instead of Softmax Cross-Entropy"),

		Arg("--from_scratch", action="store_true",
			help="Do not load any weights. Train the model from scratch"),

		Arg("--label_smoothing", type=float, default=0,
			help="Factor for label smoothing"),

		Arg("--only_head", action="store_true", help="fine-tune only last layer"),

	])\
	.seed()\
	.batch_size()\
	.epochs()\
	.debug()\
	.learning_rate(lr=1e-2, lrs=10, lrt=1e-5, lrd=1e-1)\
	.weight_decay(default=5e-4)

	parser.add_args(_args, group_name="Training arguments")

	_args = [
		Arg("--augmentations",
			choices=[
				"random_crop",
				"random_flip",
				"random_rotation",
				"center_crop",
				"color_jitter"
			],
			default=["random_crop", "random_flip", "color_jitter"],
			nargs="*"),

		Arg("--center_crop_on_val", action="store_true"),
		Arg("--brightness_jitter", type=int, default=0.3),
		Arg("--contrast_jitter", type=int, default=0.3),
		Arg("--saturation_jitter", type=int, default=0.3),

	]
	parser.add_args(_args, group_name="Augmentation arguments")

	_args = [
		Arg("--only_eval", action="store_true", help="evaluate the model only. do not train!"),
		Arg("--init_eval", action="store_true", help="evaluate the model before training"),
	]

	parser.add_args(_args, group_name="Evaluation arguments")

	_args = [
		Arg("--no_progress", action="store_true", help="dont show progress bar"),
		Arg("--no_snapshot", action="store_true", help="do not save trained model"),
		Arg("--output", "-o", type=str, default=".out", help="output folder"),
	]
	parser.add_args(_args, group_name="Output arguments")


class TrainingParserMixin(abc.ABC):
	def __init__(self, *args, **kwargs):
		super(TrainingParserMixin, self).__init__(*args, **kwargs)
		add_training_args(self)


__all__ = [
	"TrainingParserMixin",
	"add_training_args"
]



import chainer
import chainer.functions as F
import numpy as np

import abc
import logging

from chainer.backends import cuda
from chainer.optimizer_hooks import Lasso
from chainer.optimizer_hooks import WeightDecay
from chainer.serializers import save_npz
from chainer.training import extensions

from chainer_addons.functions import smoothed_cross_entropy
from chainer_addons.models import Classifier
from chainer_addons.models import ModelType
from chainer_addons.models import PrepareType
from chainer_addons.training import optimizer
from chainer_addons.training import optimizer_hooks

from cvdatasets import AnnotationType
from cvdatasets.utils import new_iterator
from cvdatasets.utils import pretty_print_dict
from cvdatasets.dataset.image import Size

from functools import partial
from os.path import join

from bdb import BdbQuit


def check_param_for_decay(param):
	return param.name != "alpha"

class _ModelMixin(abc.ABC):
	"""This mixin is responsible for optimizer creation, model creation,
	model wrapping around a classifier and model weights loading.
	"""

	def __init__(self, classifier_cls, classifier_kwargs={}, model_kwargs={}, *args, **kwargs):
		super(_ModelMixin, self).__init__(*args, **kwargs)
		self.classifier_cls = classifier_cls
		self.classifier_kwargs = classifier_kwargs
		self.model_kwargs = model_kwargs

	def wrap_model(self, opts):

		clf_class, kwargs = self.classifier_cls, self.classifier_kwargs

		self.clf = clf_class(
			model=self.model,
			loss_func=self._loss_func(opts),
			**kwargs)

		logging.info(" ".join([
			f"Wrapped the model around {clf_class.__name__}",
			f"with kwargs: {pretty_print_dict(kwargs)}",
		]))

	def _loss_func(self, opts):
		if getattr(opts, "l1_loss", False):
			return F.hinge

		elif getattr(opts, "label_smoothing", 0) >= 0:
			assert getattr(opts, "label_smoothing", 0) < 1, \
				"Label smoothing factor must be less than 1!"
			return partial(smoothed_cross_entropy,
				N=self.n_classes,
				eps=getattr(opts, "label_smoothing", 0))
		else:
			return F.softmax_cross_entropy

	def init_optimizer(self, opts):
		"""Creates an optimizer for the classifier """
		if not hasattr(opts, "optimizer"):
			self.opt = None
			return

		opt_kwargs = {}
		if opts.optimizer == "rmsprop":
			opt_kwargs["alpha"] = 0.9

		self.opt = optimizer(opts.optimizer,
			self.clf,
			opts.learning_rate,
			decay=0, gradient_clipping=False, **opt_kwargs
		)

		if opts.decay > 0:
			reg_kwargs = {}
			if opts.l1_loss:
				reg_cls = Lasso

			elif opts.pooling == "alpha":
				reg_cls = optimizer_hooks.SelectiveWeightDecay
				reg_kwargs["selection"] = check_param_for_decay

			else:
				reg_cls = WeightDecay

			logging.info(f"Adding {reg_cls.__name__} ({opts.decay:e})")
			self.opt.add_hook(reg_cls(opts.decay, **reg_kwargs))

		if getattr(opts, "only_head", False):
			assert not opts.recurrent, "FIX ME! Not supported yet!"

			logging.warning("========= Fine-tuning only classifier layer! =========")
			self.model.disable_update()
			self.model.fc.enable_update()

	def init_model(self, opts):
		"""creates backbone CNN model. This model is wrapped around the classifier later"""

		self.model = ModelType.new(
			model_type=self.model_info.class_key,
			input_size=Size(opts.input_size),
			**self.model_kwargs,
		)

	def load_model_weights(self, args):
		if getattr(args, "from_scratch", False):
			logging.info("Training a {0.__class__.__name__} model from scratch!".format(self.model))
			loader = self.model.reinitialize_clf
			self.weights = None
		else:
			if args.load:
				self.weights = args.load
				msg = "Loading already fine-tuned weights from \"{}\""
				loader_func = self.model.load_for_inference
			else:
				if args.weights:
					msg = "Loading custom pre-trained weights \"{}\""
					self.weights = args.weights

				else:
					msg = "Loading default pre-trained weights \"{}\""
					self.weights = join(
						self.data_info.BASE_DIR,
						self.data_info.MODEL_DIR,
						self.model_info.folder,
						self.model_info.weights
					)

				loader_func = self.model.load_for_finetune

			logging.info(msg.format(self.weights))
			kwargs = dict(
				weights=self.weights,
				strict=args.load_strict,
				headless=args.headless,
			)
			loader = partial(loader_func, **kwargs)

		feat_size = self.model.meta.feature_size

		if hasattr(self.clf, "output_size"):
			feat_size = self.clf.output_size

		if hasattr(self.clf, "loader"):
			loader = self.clf.loader(loader)

		logging.info(f"Part features size after encoding: {feat_size}")
		loader(n_classes=self.n_classes, feat_size=feat_size)
		self.clf.cleargrads()

class _DatasetMixin(abc.ABC):
	"""
		This mixin is responsible for annotation loading and for
		dataset and iterator creation.
	"""

	def __init__(self, dataset_cls, dataset_kwargs_factory, *args, **kwargs):
		super(_DatasetMixin, self).__init__(*args, **kwargs)
		self.dataset_cls = dataset_cls
		self.dataset_kwargs_factory = dataset_kwargs_factory

	@property
	def n_classes(self):
		return self.ds_info.n_classes + self.dataset_cls.label_shift

	def new_dataset(self, opts, size, part_size, subset):
		"""Creates a dataset for a specific subset and certain options"""
		if self.dataset_kwargs_factory is not None and callable(self.dataset_kwargs_factory):
			kwargs = self.dataset_kwargs_factory(opts, subset)
		else:
			kwargs = dict()

		kwargs.update(dict(
			subset=subset,
			dataset_cls=self.dataset_cls,
		))


		if not getattr(opts, "only_head", False):
			kwargs.update(dict(
				prepare=self.prepare,
				size=size,
				part_size=part_size,
				center_crop_on_val=getattr(opts, "center_crop_on_val", False),

			))

		ds = self.annot.new_dataset(**kwargs)
		logging.info("Loaded {} images".format(len(ds)))
		return ds

	def init_annotations(self, opts):
		"""Reads annotations and creates annotation instance, which holds important infos about the dataset"""

		self.annot = AnnotationType.new_annotation(opts, load_strict=False)

		self.data_info = self.annot.info
		self.model_info = self.data_info.MODELS[opts.model_type]
		self.ds_info = self.data_info.DATASETS[opts.dataset]
		# self.part_info = self.data_info.PART_TYPES[opts.parts]

		if getattr(opts, "only_head", False):
			self.annot.feature_model = opts.model_type

		self.dataset_cls.label_shift = opts.label_shift


	def init_datasets(self, opts):

		size = Size(opts.input_size)
		part_size = getattr(opts, "parts_input_size", None)
		part_size = size if part_size is None else Size(part_size)

		self.prepare = partial(PrepareType[opts.prepare_type](self.model),
			swap_channels=opts.swap_channels,
			keep_ratio=getattr(opts, "center_crop_on_val", False),
		)

		logging.info(" ".join([
			f"Created {self.model.__class__.__name__} model",
			f"with \"{opts.prepare_type}\" prepare function."
		]))

		logging.info(" ".join([
			f"Image input size: {size}",
			f"Image parts input size: {part_size}",
		]))

		self.train_data = self.new_dataset(opts, size, part_size, "train")
		self.val_data = self.new_dataset(opts, size, part_size, "test")

	def init_iterators(self, opts):
		"""Creates training and validation iterators from training and validation datasets"""

		kwargs = dict(n_jobs=opts.n_jobs, batch_size=opts.batch_size)

		if hasattr(self.train_data, "new_iterator"):
			self.train_iter, _ = self.train_data.new_iterator(**kwargs)
		else:
			self.train_iter, _ = new_iterator(self.train_data, **kwargs)

		if hasattr(self.val_data, "new_iterator"):
			self.val_iter, _ = self.val_data.new_iterator(**kwargs,
				repeat=False, shuffle=False
			)
		else:
			self.val_iter, _ = new_iterator(self.val_data,
				**kwargs, repeat=False, shuffle=False
			)


class _TrainerMixin(abc.ABC):
	"""This mixin is responsible for updater, evaluator and trainer creation.
	Furthermore, it implements the run method
	"""

	def __init__(self, updater_cls, updater_kwargs={}, *args, **kwargs):
		super(_TrainerMixin, self).__init__(*args, **kwargs)
		self.updater_cls = updater_cls
		self.updater_kwargs = updater_kwargs

	def init_updater(self):
		"""Creates an updater from training iterator and the optimizer."""

		if self.opt is None:
			self.updater = None
			return

		self.updater = self.updater_cls(
			iterator=self.train_iter,
			optimizer=self.opt,
			device=self.device,
			**self.updater_kwargs,
		)
		logging.info(" ".join([
			f"Using single GPU: {self.device}.",
			f"{self.updater_cls.__name__} is initialized",
			f"with following kwargs: {pretty_print_dict(self.updater_kwargs)}"
			])
		)

	def init_evaluator(self, default_name="val"):
		"""Creates evaluation extension from validation iterator and the classifier."""

		self.evaluator = extensions.Evaluator(
			iterator=self.val_iter,
			target=self.clf,
			device=self.device,
			progress_bar=True
		)

		self.evaluator.default_name = default_name

	def run(self, trainer_cls, opts, *args, **kwargs):

		trainer = trainer_cls(
			opts=opts,
			updater=self.updater,
			evaluator=self.evaluator,
			weights=self.weights,
			*args, **kwargs
		)

		logging.info("Snapshotting is {}abled".format("dis" if opts.no_snapshot else "en"))

		def dump(suffix):
			if opts.only_eval or opts.no_snapshot:
				return

			save_npz(join(trainer.out,
				"clf_{}.npz".format(suffix)), self.clf)
			save_npz(join(trainer.out,
				"model_{}.npz".format(suffix)), self.model)

		try:
			trainer.run(opts.init_eval or opts.only_eval)
		except (KeyboardInterrupt, BdbQuit) as e:
			raise e
		except Exception as e:
			dump("exception")
			raise e
		else:
			dump("final")


class DefaultFinetuner(_ModelMixin, _DatasetMixin, _TrainerMixin):
	""" The default Finetuner gathers together the creations of all needed
	components and call them in the correct order

	"""

	def __init__(self, opts, *args, **kwargs):
		super(DefaultFinetuner, self).__init__(*args, **kwargs)

		self.gpu_config(opts)
		cuda.get_device_from_id(self.device).use()

		self.init_annotations(opts)
		self.init_model(opts)

		self.init_datasets(opts)
		self.init_iterators(opts)

		self.wrap_model(opts)
		self.load_model_weights(opts)

		self.init_optimizer(opts)
		self.init_updater()
		self.init_evaluator()

	def gpu_config(self, opts):
		if -1 in opts.gpu:
			self.device = -1
		else:
			self.device = opts.gpu[0]




import abc

from cvargparse import Arg
from cvfinetune.parser.utils import get_info_file
from cvfinetune.parser.utils import parser_extender

from chainer_addons.links import PoolingType
from chainer_addons.models import PrepareType

@parser_extender
def add_model_args(parser):

	info_file = get_info_file()

	if info_file is None:
		model_type_choices = None
	else:
		model_type_choices = info_file.MODELS.keys()

	_args = [
		Arg("--model_type", "-mt",
			default="resnet", choices=model_type_choices,
			help="type of the model"),

		Arg("--input_size", type=int, nargs="+", default=0,
			help="overrides default input size of the model, if greater than 0"),

		Arg("--parts_input_size", type=int, nargs="+", default=0,
			help="overrides default input part size of the model, if greater than 0"),

		PrepareType.as_arg("prepare_type",
			help_text="type of image preprocessing"),

		PoolingType.as_arg("pooling",
			help_text="type of pre-classification pooling"),

		Arg("--load", type=str,
			help="ignore weights and load already fine-tuned model (classifier will NOT be re-initialized and number of classes will be unchanged)"),

		Arg("--weights", type=str,
			help="ignore default weights and load already pre-trained model (classifier will be re-initialized and number of classes will be changed)"),

		Arg("--headless", action="store_true",
			help="ignores classifier layer during loading"),

		Arg("--load_strict", action="store_true",
			help="load weights in a strict mode"),
	]

	parser.add_args(_args, group_name="Model arguments")


class ModelParserMixin(abc.ABC):
	def __init__(self, *args, **kwargs):
		super(ModelParserMixin, self).__init__(*args, **kwargs)
		add_model_args(self)


__all__ = [
	"ModelParserMixin",
	"add_model_args"
]




import logging
try:
	import chainermn
except Exception as e: #pragma: no cover
	_CHAINERMN_AVAILABLE = False #pragma: no cover
else:
	_CHAINERMN_AVAILABLE = True

from cvfinetune import utils
from cvfinetune.finetuner.base import DefaultFinetuner
from cvfinetune.finetuner.mpi import MPIFinetuner

from cvdatasets.utils import pretty_print_dict

class FinetunerFactory(object):

	@classmethod
	def new(cls, opts, default=DefaultFinetuner, mpi_tuner=MPIFinetuner):

		if getattr(opts, "mpi", False):
			assert _CHAINERMN_AVAILABLE, "Distributed training is not possible!"
			msg1 = "MPI enabled. Creating NCCL communicator!"
			comm = chainermn.create_communicator("pure_nccl")
			msg2 = f"Rank: {comm.rank}, IntraRank: {comm.intra_rank}, InterRank: {comm.inter_rank}"

			utils.log_messages([msg1, msg2])
			return cls(mpi_tuner, comm=comm)
		else:
			return cls(default)

	def __init__(self, tuner_cls, **kwargs):
		super(FinetunerFactory, self).__init__()

		self.tuner_cls = tuner_cls
		self.kwargs = kwargs
		logging.info(f"Using {self.tuner_cls.__name__} with arguments: {pretty_print_dict(self.kwargs)}")

	def __call__(self, **kwargs):
		_kwargs = dict(self.kwargs)
		_kwargs.update(kwargs)

		return self.tuner_cls(**_kwargs)

	def get(self, key, default=None):
		return self.kwargs.get(key, default)

	def __getitem__(self, key):
		return self.kwargs[key]

	def __setitem__(self, key, value):
		self.kwargs[key] = value

__all__ = [
	"get_finetuner",
	"DefaultFinetuner",
	"MPIFinetuner",
]



import logging

def log_messages(msg_list, *, n_chars=10, char="=", logger=None, level=logging.INFO):
	"""
		Adds <chars> and a space at each end of every message.
		Adjusts the length of each message to the maximum length
		of the messages in the list.
	"""

	max_len = max(map(len, msg_list))
	fmt_len = max_len + 2*(n_chars+1)
	fmt = "{:" + char + "^" + str(fmt_len) + "s}"

	logger = logger or logging.getLogger()

	for msg in msg_list:
		logger.log(level, fmt.format(f" {msg} "))



import os
import warnings

from cvargparse import BaseParser
from cvdatasets.utils import read_info_file
from functools import wraps


WARNING = """Could not find default info file \"{}\". """ + \
"""Some arguments (dataset, parts etc.) are not restraint to certain choices! """ + \
"""You can set <DATA> environment variable to change the default info file location."""

DEFAULT_INFO_FILE = os.environ.get("DATA")
def get_info_file():

	if DEFAULT_INFO_FILE is not None and os.path.isfile(DEFAULT_INFO_FILE):
		return read_info_file(DEFAULT_INFO_FILE)
	else:
		warnings.warn(WARNING.format(DEFAULT_INFO_FILE))
		return None


def parser_extender(extender):

	@wraps(extender)
	def inner(parser):
		assert isinstance(parser, BaseParser), \
			"Parser should be an BaseParser instance!"

		extender(parser)

		return parser

	return inner



from cvfinetune.training.trainer.base import Trainer, default_intervals
from chainer_addons.training.sacred import SacredTrainerMixin

class SacredTrainer(SacredTrainerMixin, Trainer):

	def __init__(self, intervals=default_intervals, *args, **kwargs):
		super(SacredTrainer, self).__init__(
			intervals=intervals, sacred_trigger=intervals.log, *args, **kwargs)



import logging
import os
import platform
import warnings


from cvargparse import Arg
from cvargparse import ArgFactory
from cvargparse import GPUParser
from cvargparse.utils import logger_config

from cvfinetune.parser.dataset_args import add_dataset_args
from cvfinetune.parser.model_args import add_model_args
from cvfinetune.parser.training_args import add_training_args


def default_factory(extra_list=[]):
	return ArgFactory(extra_list)


class FineTuneParser(GPUParser):

	def _logging_config(self, simple=False):
		if not self.has_logging: return
		fmt = '{levelname:s} - [{asctime:s}] {filename:s}:{lineno:d} [{funcName:s}]: {message:s}'

		handler0 = logging.StreamHandler()
		handler0.addFilter(HostnameFilter())
		fmt0 = "<{hostname:^10s}>: " + fmt

		if self._args.logfile is None:
			filename = f"{platform.node()}.log"
		else:
			filename = self._args.logfile

		self._file_handler = handler1 = logging.FileHandler(filename=filename, mode="w")

		logger_config.init_logging_handlers([
			(handler0, fmt0, logging.INFO),
			(handler1, fmt, logging.INFO),
		])


	def __del__(self):
		try:
			if getattr(self, "_file_handler", None) is not None:
				self._file_handler.flush()
		except Exception as e:
			warnings.warn("Could not flush logs to file: {}".format(e))


	def __init__(self, *args, **kwargs):
		super(FineTuneParser, self).__init__(*args, **kwargs)
		self._file_handler = None

		add_dataset_args(self)
		add_model_args(self)
		add_training_args(self)


class HostnameFilter(logging.Filter):

	def filter(self, record):
		record.hostname = platform.node()
		return True



import abc
import chainer
import chainer.functions as F
import chainer.links as L
import logging

from chainer_addons.models.classifier import Classifier as C

class Classifier(C):

	def __init__(self, *args, **kwargs):
		super(Classifier, self).__init__(*args, **kwargs)

		assert hasattr(self, "model"), \
			"This classifiert has no \"model\" attribute!"

	@property
	def feat_size(self):
		if hasattr(self.model.pool, "output_dim") and self.model.pool.output_dim is not None:
			return self.model.pool.output_dim

		return self.model.meta.feature_size

	@property
	def output_size(self):
		return self.feat_size


class SeparateModelClassifier(Classifier):
	"""Classifier, that holds two separate models"""

	def __init__(self, *args, **kwargs):
		super(SeparateModelClassifier, self).__init__(*args, **kwargs)

		with self.init_scope():
			self.init_separate_model()

	@abc.abstractmethod
	def __call__(self, *args, **kwargs):
		super(SeparateModelClassifier, self).__call__(*args, **kwargs)

	def init_separate_model(self):

		if hasattr(self, "separate_model"):
			logging.warn("Global Model already initialized! Skipping further execution!")
			return

		self.separate_model = self.model.copy(mode="copy")

	def loader(self, model_loader):

		def inner(n_classes, feat_size):
			# use the given feature size here
			model_loader(n_classes=n_classes, feat_size=feat_size)

			# use the given feature size first ...
			self.separate_model.reinitialize_clf(
				n_classes=n_classes,
				feat_size=feat_size)

			# then copy model params ...
			self.separate_model.copyparams(self.model)

			# now use the default feature size to re-init the classifier
			self.separate_model.reinitialize_clf(
				n_classes=n_classes,
				feat_size=self.feat_size)

		return inner



import chainermn
import logging
from chainermn import scatter_dataset as scatter

from cvfinetune.finetuner.base import DefaultFinetuner

class MPIFinetuner(DefaultFinetuner):

	def __init__(self, opts, *args, comm, **kwargs):
		self.comm = comm
		super(MPIFinetuner, self).__init__(opts, *args, **kwargs)

	@property
	def mpi(self):
		return self.comm is not None

	@property
	def mpi_main_process(self):
		return not (self.comm is not None and self.comm.rank != 0)

	def gpu_config(self, opts):
		super(MPIFinetuner, self).gpu_config(opts)

		if self.mpi:
			if len(opts.gpu) > 1:
				self.device = opts.gpu[self.comm.rank]
			else:
				self.device += self.comm.intra_rank
			ranks = f"{self.comm.rank}|{self.comm.intra_rank}|{self.comm.inter_rank}"
			logging.info(f"Node with ranks {ranks} assigned to GPU #{self.device}")
		else:
			logging.warn("Using MPIFinetuner without setting a communicator!")

	def scatter_datasets(self):
		if self.mpi:
			self.train_data = scatter(self.train_data, self.comm)
			self.val_data = scatter(self.val_data, self.comm)
		else:
			logging.warn("Data scattering was not Possible!")


	def init_datasets(self, *args, **kwargs):

		if self.mpi_main_process:
			super(MPIFinetuner, self).init_datasets(*args, **kwargs)
		else:
			self.train_data, self.val_data = None, None

		self.scatter_datasets()

	def init_optimizer(self, opts):
		super(MPIFinetuner, self).init_optimizer(opts)

		if self.mpi:
			self.opt = chainermn.create_multi_node_optimizer(self.opt, self.comm)

	def init_evaluator(self):
		super(MPIFinetuner, self).init_evaluator()

		if self.mpi:
			self.evaluator = chainermn.create_multi_node_evaluator(
				self.evaluator, self.comm)

	def run(self, trainer_cls, opts, *args, **kwargs):
		kwargs["no_observe"] = not self.mpi_main_process
		opts.no_snapshot = not self.mpi_main_process
		opts.no_progress = opts.no_progress or not self.mpi_main_process
		super(MPIFinetuner, self).run(trainer_cls, opts, *args, **kwargs)



#!/usr/bin/env python

import os
import pkg_resources
import sys

from setuptools import setup, find_packages

try: # for pip >= 10
	from pip._internal.req import parse_requirements
except ImportError: # for pip <= 9.0.3
	from pip.req import parse_requirements

import cvfinetune
install_requires = [line.strip() for line in open("requirements.txt").readlines()]


setup(
	name='cvfinetune',
	python_requires=">3.5",
	version=cvfinetune.__version__,
	description='Fine-tune framework based on chainer',
	author='Dimitri Korsch',
	author_email='korschdima@gmail.com',
	license='MIT License',
	packages=find_packages(),
	zip_safe=False,
	setup_requires=[],
	install_requires=install_requires,
    package_data={'': ['requirements.txt']},
    data_files=[('.',['requirements.txt'])],
    include_package_data=True,
)



import pymongo as pym
import numpy as np
import matplotlib.pyplot as plt
import warnings
from datetime import datetime

from matplotlib.gridspec import GridSpec

def parse_date(date, fmt="{:d}-{:02d}-{:02d} 00:00:00"):
	if isinstance(date, str):
		return datetime.fromisoformat(date)

	elif isinstance(date, (tuple, list)):
		return datetime.fromisoformat(fmt.format(*date))

	elif date is None:
		return None

	else:
		warnings.warn("Unsupported date format: {date}. Date parsing skipped".format(date))

def query_to_str(query):
	query_list = [f"{key}={value}" for key, value in query.items()]
	return "{{{}}}".format(", ".join(query_list))

class SacredPlotter(object):

	@staticmethod
	def auth_url(creds, host="localhost"):
		url = "mongodb://{user}:{password}@{host}:27017/{db_name}?authSource=admin".format(
			host=host, **creds)
		return url

	def __init__(self, creds):
		super(SacredPlotter, self).__init__()
		self.client = pym.MongoClient(SacredPlotter.auth_url(creds))

		self.db = self.client[creds["db_name"]]
		self.metrics = self.db["metrics"]
		self.runs = self.db["runs"]

	def get_values(self, metric, query):

		values = []
		runs = list(self.runs.find(query))

		for run in runs:

			accuracies = self.metrics.find_one(dict(
				name=metric,
				run_id=run["_id"],
			))
			if accuracies is None or accuracies["values"] is None:
				continue

			values.append(accuracies["values"][-1])

		return values

	def plot(self,
		metrics,
		setups,
		query_factory,
		setup_to_label,
		include_running=False,
		start_time=None,
		metrics_key="val/main/",
		**plot_kwargs):

		"""
			Arguments:
				- metrics: defines which metrics to plot

				- setups: defines different setups, that
					will be compared

				- query_factory: callable; creates from the setup
					specification the pymongo query

				- setup_to_label: callable; converts a setup
					specification into a readable label

				- include_running: whether running experiments should
					be included or not

				- start_time: time (starting of an experiment) from which
					the experiments should be considered

				- metrics_key: prefix that will be appended to each
					metric name

		"""

		n_metrics = len(metrics)
		n_cols = int(np.ceil(np.sqrt(n_metrics)))
		n_rows = int(np.ceil(n_metrics / n_cols))

		grid = GridSpec(n_rows, n_cols)
		axs = []
		for i, metric in enumerate(metrics):

			res = []

			for setup in setups:
				query = query_factory(setup)

				if not include_running:
					query["status"] = {"$ne": "RUNNING"}

				if start_time is not None:
					query["start_time"] = {"$gte": parse_date(start_time)}

				values = self.get_values(f"{metrics_key}{metric}", query)
				if len(values) == 0:
					warnings.warn(f"No values found for query { query_to_str(query) }")
				else:
					res.append((setup, values))

			row, col = np.unravel_index(i, (n_rows, n_cols))

			if len(res) == 0:
				warnings.warn(f"No setups for metric \"{metric}\" collected!")

			else:

				ax = plt.subplot(grid[row, col])
				labels, values = zip(*[(setup_to_label(setup, vals), vals) for setup, vals in res if vals])
				ax.boxplot(values, labels=labels, **plot_kwargs)
				ax.set_title(f"Metric: {metric}")

				axs.append(ax)

		return axs





import logging
from os.path import join, basename
from datetime import datetime

import chainer
from chainer.training import extensions, Trainer as T
from chainer.training import trigger as trigger_module
from chainer_addons.training import lr_shift
from chainer_addons.training.optimizer import OptimizerType
from chainer_addons.training.extensions.learning_rate import CosineAnnealingLearningRate
from chainer_addons.training.extensions import AlternateTrainable, SwitchTrainables, WarmUp

from cvdatasets.utils import attr_dict

default_intervals = attr_dict(
	print =		(1,  'epoch'),
	log =		(1,  'epoch'),
	eval =		(1,  'epoch'),
	snapshot =	(10, 'epoch'),
)

def debug_hook(trainer):
	pass
	# print(trainer.updater.get_optimizer("main").target.model.fc6.W.data.mean(), file=open("debug.out", "a"))


def _is_adam(opts):
	return opts.optimizer == OptimizerType.ADAM.name.lower()

class Trainer(T):
	_default_base_model = "model"

	def __init__(self, opts, updater, evaluator=None, weights=None, intervals=default_intervals, no_observe=False):

		self._only_eval = opts.only_eval
		if weights is None or weights == "auto":
			self.base_model = self._default_base_model
		else:
			self.base_model, _, _ = basename(weights).rpartition(".")

		optimizer = updater.get_optimizer("main")
		# adam has some specific attributes, so we need to check this
		is_adam = _is_adam(opts)
		clf = optimizer.target
		model = clf.model

		if no_observe:
			outdir = opts.output
		else:
			outdir = self.output_directory(opts)
			logging.info("Training outputs are saved under \"{}\"".format(outdir))

		super(Trainer, self).__init__(
			updater=updater,
			stop_trigger=(opts.epochs, 'epoch'),
			out=outdir
		)

		### Evaluator ###
		self.evaluator = evaluator
		if evaluator is not None:
			self.extend(evaluator, trigger=intervals.eval)

		### Warm up ###
		lr_offset = 0
		if opts.warm_up:
			assert opts.warm_up > 0, "Warm-up argument must be positive!"
			lr_offset = opts.warm_up

			warm_up_lr = opts.learning_rate
			logging.info("Warm-up of {} epochs enabled!".format(opts.warm_up))
			self.extend(WarmUp(
				opts.warm_up, model,
				opts.learning_rate, warm_up_lr))


		### LR shift ###
		if opts.cosine_schedule is not None and opts.cosine_schedule > 0:
			lr_shift_ext = CosineAnnealingLearningRate(
				attr="alpha" if is_adam else "lr",
				lr=opts.learning_rate,
				target=opts.lr_target,
				epochs=opts.epochs,
				offset=lr_offset,
				stages=opts.cosine_schedule
			)
			new_epochs = lr_shift_ext._epochs
			self.stop_trigger = trigger_module.get_trigger((new_epochs, "epoch"))
			self.extend(lr_shift_ext)
		else:
			lr_shift_ext = lr_shift(optimizer,
				init=opts.learning_rate,
				rate=opts.lr_decrease_rate, target=opts.lr_target)
			self.extend(lr_shift_ext, trigger=(opts.lr_shift, 'epoch'))

		### Code below is only for "main" Trainers ###
		if no_observe: return

		self.extend(extensions.observe_lr(), trigger=intervals.log)
		self.extend(extensions.LogReport(trigger=intervals.log))

		### Snapshotting ###
		self.setup_snapshots(opts, clf.model, intervals.snapshot)

		### Reports and Plots ###
		print_values, plot_values = self.reportables(opts)
		self.extend(extensions.PrintReport(print_values), trigger=intervals.print)
		for name, values in plot_values.items():
			ext = extensions.PlotReport(values, 'epoch', file_name='{}.png'.format(name))
			self.extend(ext)

		### Progress bar ###
		if not opts.no_progress:
			self.extend(extensions.ProgressBar(update_interval=1))

	def setup_snapshots(self, opts, obj, trigger):

		if opts.no_snapshot:
			logging.warning("Models are not snapshot!")
		else:
			dump_fmt = "ft_model_epoch{0.updater.epoch:03d}.npz"
			self.extend(extensions.snapshot_object(obj, dump_fmt), trigger=trigger)
			logging.info("Snapshot format: \"{}\"".format(dump_fmt))

	def eval_name(self, name):
		if self.evaluator is None:
			return name

		return f"{self.evaluator.default_name}/{name}"

	def reportables(self, opts):

		print_values = [
			"elapsed_time",
			"epoch",
			# "lr",

			"main/accuracy", self.eval_name("main/accuracy"),
			"main/loss", self.eval_name("main/loss"),

		]

		plot_values = {
			"accuracy": [
				"main/accuracy",  self.eval_name("main/accuracy"),
			],
			"loss": [
				"main/loss", self.eval_name("main/loss"),
			],
		}

		return print_values, plot_values


	def output_directory(self, opts):

		result = opts.output

		if self.base_model != self._default_base_model:
			result = join(result, self.base_model)

		result = join(result, datetime.now().strftime("%Y-%m-%d-%H.%M.%S.%f"))
		return result

	def run(self, init_eval=True):
		if init_eval:
			logging.info("Evaluating initial model ...")
			init_perf = self.evaluator(self)
			values = {key: float(value) for key, value in init_perf.items()}

			msg = []

			if "val/main/accuracy" in values:
				msg.append("Initial accuracy: {val/main/accuracy:.3%}".format(**values))

			if "val/main/loss" in values:
				msg.append("Initial loss: {val/main/loss:.3f}".format(**values))

			logging.info(" ".join(msg))

		if self._only_eval:
			return
		return super(Trainer, self).run()

